/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "houdini/houdini.pmh"

#define	FE_HSOP_DISCOVER_DEBUG	FALSE
#define	FE_HSOP_CATALOG_DEBUG	FALSE
#define	FE_HSOP_COOK_DEBUG		FALSE
#define	FE_HSOP_SHOW_CLASS		FALSE

extern "C"
{

//* discovery point for Houdini
FE_DL_EXPORT void newSopOperator(OP_OperatorTable *table)
{
	char* buffer=new char[123];
	delete[] buffer;
	fe::HoudiniSOP::discoverOperators(table);
}

}

namespace fe
{

static PRM_Name	gs_sopNames[] = {
    PRM_Name("fe_component", "FE Component"),
    PRM_Name("reload", "Reload")
};

static PRM_Template gs_houdiniTemplateList[] =
{
    PRM_Template(PRM_STRING, 1, &gs_sopNames[0], 0),
    PRM_Template(PRM_CALLBACK, 1, &gs_sopNames[1], 0, 0, 0,
			&HoudiniSOP::reloadCallback),
	PRM_Template()
};

int HoudiniSOP::reloadCallback(void *pData,int index, float time,
	const PRM_Template* pTemplate)
{
	feLog("HoudiniSOP::reloadCallback\n");

	HoudiniSOP* pHoudiniSOP = (HoudiniSOP*)pData;
	pHoudiniSOP->m_lastName="";

	feLog("HoudiniSOP::reloadCallback done\n");

	return 1;
}

int HoudiniSOP::discoverOperators(OP_OperatorTable *table)
{
#if	FE_HSOP_DISCOVER_DEBUG
	feLog("HoudiniSOP:discoverOperators\n");
#endif

	HoudiniContext houdiniContext;

#if	FE_HSOP_DISCOVER_DEBUG
	feLog("HoudiniSOP:discoverOperators master=%p\n",
		houdiniContext.master().raw());
#endif

	sp<Registry> spRegistry=houdiniContext.master()->registry();
	houdiniContext.loadLibraries();

	int count=0;

	std::vector<String> available;
	spRegistry->listAvailable("OperatorSurfaceI",available);
	U32 candidates=available.size();
	for(U32 index=0;index<candidates;index++)
	{
#if	FE_HSOP_DISCOVER_DEBUG
		feLog("checking %d/%d \"%s\"\n",
				index,candidates,available[index].c_str());
#endif

		sp<OperatorSurfaceI> spOperatorSurfaceI=
				spRegistry->create(available[index]);
		if(!spOperatorSurfaceI.isValid())
		{
			feLog("HoudiniSOP:discoverOperators"
					" Component is not an OperatorSurfaceI \"%s\"\n",
					available[index].c_str());
			continue;
		}

		//* ignore operators tagged as debug
		if(available[index].dotMatch("*.*.*.debug"))
		{
#if FE_CODEGEN<=FE_DEBUG
			feLog("HoudiniSOP:discoverOperators allowing debug operator:\n"
					"  %s\n",available[index].c_str());
#else
			continue;
#endif
		}

		//* ignore operators tagged as nohoudini
		if(available[index].dotMatch("*.*.*.nohoudini"))
		{
#if FE_CODEGEN<=FE_DEBUG
			feLog("HoudiniSOP:discoverOperators ignoring operator:\n"
					"  %s\n",available[index].c_str());
#endif
			continue;
		}

		String feName=available[index];
		char* buffer1=strchr(const_cast<char*>(available[index].c_str()),'.');
		if(buffer1)
		{
			buffer1++;
			const char* buffer2=strchr(buffer1,'.');
			if(buffer2>buffer1)
			{
				feName.sPrintf("%.*s",int(buffer2-buffer1),buffer1);
			}
			else
			{
				feName=buffer1;
			}
		}
		String shortName=feName;
		if(!strncmp(feName.c_str(),"Operator",8))
		{
			shortName=feName.c_str()+8;
		}

#if	FE_HSOP_DISCOVER_DEBUG
		feLog("HoudiniSOP::discoverOperators feName \"%s\" shortName \"%s\"\n",
				feName.c_str(), shortName.c_str());
#endif

		std::vector<String> keys;

		sp<Catalog> spCatalog=spOperatorSurfaceI;
		if(!spCatalog.isValid())
		{
#if FE_CODEGEN<=FE_DEBUG
			feLog("HoudiniSOP:discoverOperators"
					" Component is not a Catalog \"%s\"\n",
					available[index].c_str());
#endif
		}
		else
		{
			spCatalog->catalogKeys(keys);
		}

		const U32 number=keys.size();

#if	FE_HSOP_DISCOVER_DEBUG
		feLog("HoudiniSOP::discoverOperators keys %d\n",number);
#endif

		const BWORD showClass=FE_HSOP_SHOW_CLASS;
		const U32 given=showClass? 2: 1;

		PRM_Template* pTemplateList=new PRM_Template[given+number+1];
		pTemplateList[0]=gs_houdiniTemplateList[0];

		if(showClass)
		{
			pTemplateList[1]=gs_houdiniTemplateList[1];
		}
		else
		{
			pTemplateList[0].setInvisible(true);
		}

		pTemplateList[0].setStringDefault(0,available[index].c_str());

		PRM_Name* pNameList=new PRM_Name[number];

		char** inputlabels=NULL;
		unsigned minSources=0;
		unsigned maxSources=0;
		U32 templateIndex=given;

		for(U32 m=0;m<number;m++)
		{
#if	FE_HSOP_DISCOVER_DEBUG
			feLog("HoudiniSOP::discoverOperators key %d: \"%s\"\n",
					m,keys[m].c_str());
#endif

			Instance instance;
			bool success=spCatalog->catalogLookup(keys[m],instance);
			if(!success)
			{
				feLog("HoudiniSOP::discoverOperators invalid Instance\n");
				continue;
			}

			//* leaks?
			const char* cleanname=strdup(keys[m].substitute(" ","_").c_str());
			const char* keyname=strdup(keys[m].c_str());
			const char* labelName=strdup(spCatalog->catalogOrDefault<String>(
					keyname,"label",keyname).c_str());
			const String helpText=
					spCatalog->catalogOrDefault<String>(keyname,"hint","");

			pNameList[m]=PRM_Name(cleanname,labelName);

			if(instance.is<String>())
			{
				char* value=strdup(instance.cast<String>().c_str());

				pTemplateList[templateIndex]=
						PRM_Template(PRM_STRING,1,&pNameList[m],
						NULL,NULL,NULL,NULL,NULL,1,helpText.c_str());
				pTemplateList[templateIndex].setStringDefault(0,value);
				templateIndex++;
			}
			else if(instance.is<bool>())
			{
				bool value=instance.cast<bool>();

				pTemplateList[templateIndex]=
						PRM_Template(PRM_TOGGLE,1,&pNameList[m],
						NULL,NULL,NULL,NULL,NULL,1,helpText.c_str());
				pTemplateList[templateIndex].setOrdinalDefault(0,value);
				templateIndex++;
			}
			else if(instance.is<Real>())
			{
				Real value=instance.cast<Real>();

				//* Unfortunately, it appears that we can't set the
				//*	UI limits separate from the PRM limits.

				Real min=spCatalog->catalogOrDefault<Real>(keyname,"min",0.0);
				Real max=spCatalog->catalogOrDefault<Real>(keyname,"max",1.0);
				Real low=spCatalog->catalogOrDefault<Real>(keyname,"low",min);
				Real high=spCatalog->catalogOrDefault<Real>(keyname,"high",max);

				PRM_RangeFlag prmRangeFlagMin=
						(low==min)? PRM_RANGE_RESTRICTED: PRM_RANGE_UI;
				PRM_RangeFlag prmRangeFlagMax=
						(high==max)? PRM_RANGE_RESTRICTED: PRM_RANGE_UI;

				PRM_Range* pPrmRange=new PRM_Range(prmRangeFlagMin,low,
						prmRangeFlagMax,high);
				pTemplateList[templateIndex]=
						PRM_Template(PRM_FLT,1,&pNameList[m],
						NULL,NULL,pPrmRange,NULL,NULL,1,helpText.c_str());
				pTemplateList[templateIndex].setFloatDefault(0,value);
				templateIndex++;
			}
			else if(instance.is<I32>())
			{
				I32 value=instance.cast<I32>();

				I32 min=spCatalog->catalogOrDefault<I32>(keyname,"min",0);
				I32 max=spCatalog->catalogOrDefault<I32>(keyname,"max",10);
				I32 low=spCatalog->catalogOrDefault<I32>(keyname,"low",min);
				I32 high=spCatalog->catalogOrDefault<I32>(keyname,"high",max);

				PRM_RangeFlag prmRangeFlagMin=
						(low==min)? PRM_RANGE_RESTRICTED: PRM_RANGE_UI;
				PRM_RangeFlag prmRangeFlagMax=
						(high==max)? PRM_RANGE_RESTRICTED: PRM_RANGE_UI;

				PRM_Range* pPrmRange=new PRM_Range(prmRangeFlagMin,low,
						prmRangeFlagMax,high);
				pTemplateList[templateIndex]=
						PRM_Template(PRM_INT,1,&pNameList[m],
						NULL,NULL,pPrmRange,NULL,NULL,1,helpText.c_str());
				pTemplateList[templateIndex].setOrdinalDefault(0,value);
				templateIndex++;
			}
			else if(instance.is<SpatialVector>())
			{
				SpatialVector& value=instance.cast<SpatialVector>();

				pTemplateList[templateIndex]=
						PRM_Template(PRM_XYZ,3,&pNameList[m],
						NULL,NULL,NULL,NULL,NULL,1,helpText.c_str());
				pTemplateList[templateIndex].setFloatDefault(0,value[0]);
				pTemplateList[templateIndex].setFloatDefault(1,value[1]);
				pTemplateList[templateIndex].setFloatDefault(2,value[2]);
				templateIndex++;
			}
			else if(instance.is<Vector4>())
			{
				Vector4& value=instance.cast<Vector4>();

				pTemplateList[templateIndex]=
						PRM_Template(PRM_FLT,4,&pNameList[m],
						NULL,NULL,NULL,NULL,NULL,1,helpText.c_str());
				pTemplateList[templateIndex].setFloatDefault(0,value[0]);
				pTemplateList[templateIndex].setFloatDefault(1,value[1]);
				pTemplateList[templateIndex].setFloatDefault(2,value[2]);
				pTemplateList[templateIndex].setFloatDefault(3,value[3]);
				templateIndex++;
			}
			else if(instance.is< sp<Component> >())
			{
				const String io=spCatalog->catalogOrDefault<String>(keyname,
						"IO","input");
				if(io!="input")
				{
					continue;
				}

				const String implementation=
						spCatalog->catalogOrDefault<String>(keyname,
						"implementation",
						"SurfaceAccessibleI.SurfaceAccessibleHoudini");

				if(implementation=="SurfaceAccessibleI" ||
						implementation==
						"SurfaceAccessibleI.SurfaceAccessibleHoudini")
				{
					const bool optional=spCatalog->catalogOrDefault<bool>(
							keyname,"optional",false);

					if(!optional)
					{
						minSources++;
					}
					maxSources++;

					char** newlabels=new char*[maxSources+1];
					for(U32 m=0;m<maxSources-1;m++)
					{
						newlabels[m]=inputlabels[m];
					}
					newlabels[maxSources-1]=strdup(labelName);
					newlabels[maxSources]=NULL;

					delete[] inputlabels;
					inputlabels=newlabels;
				}
			}
		}

		pTemplateList[templateIndex]=PRM_Template();

		const char* name=strdup(feName.c_str());
		const char* english=strdup(("FE "+shortName).c_str());
		CH_LocalVariable* variables=NULL;
		const unsigned flags=0x0;

#if	FE_HSOP_DISCOVER_DEBUG
		feLog("HoudiniSOP::discoverOperators %s sources %d-%d\n",
				name,minSources,maxSources);
#endif

		table->addOperator(new OP_Operator(
				name,
				english,
				fe::HoudiniSOP::create,
				pTemplateList,
				minSources,
				maxSources,
				variables,
				flags,
				(const char**)inputlabels));

//		LEAK or released by Houdini?
//		delete[] pNameList;
//		delete[] pTemplateList;

		count++;
	}

	return count;
}

HoudiniSOP::HoudiniSOP(OP_Network *net, const char *name, OP_Operator *op):
	m_pOpOperator(op),
	SOP_Node(net, name, op),
	m_lastTime(0.0f),
	m_pGdp(NULL)
{
	op->setIconName("FE");

//	feLog("HoudiniSOP master=%p\n",m_houdiniContext.master().raw());

	sp<Registry> spRegistry=m_houdiniContext.master()->registry();
	m_houdiniContext.loadLibraries();

	m_spDrawI=spRegistry->create("DrawI.HoudiniDraw");
//	feLog("HoudiniSOP m_spDrawI valid=%d\n",m_spDrawI.isValid());

	sp<Scope> spScope=m_houdiniContext.scope();

	m_spSignalerI=spRegistry->create("SignalerI");
	m_spSignalerI->setName("HoudiniSOP.SignalerI");

	m_aSurfaceOutput.setup(spScope,"SurfaceOutput","component");
	m_aDrawI.setup(spScope,"DrawI","component");

	m_spLayout = spScope->declare("SOP");
	m_spLayout->populate(m_aSurfaceOutput);
	m_spLayout->populate(m_aDrawI);
	m_signal = m_houdiniContext.scope()->createRecord(m_spLayout);
	m_aDrawI(m_signal)=m_spDrawI;

#if FE_COUNTED_TRACK
	Counted::registerRegion(this,sizeof(HoudiniSOP),"HoudiniSOP");
#endif
}

HoudiniSOP::~HoudiniSOP(void)
{
	feLog("~HoudiniSOP\n");

#if FE_COUNTED_TRACK
	Counted::deregisterRegion(this);
#endif

	m_pOpOperator=NULL;
}

void HoudiniSOP::updateCatalog(Real time)
{
	if(!m_spOperatorSurfaceI.isValid())
	{
		feLog("HoudiniSOP::updateCatalog invalid m_spOperatorSurfaceI\n");
		return;
	}

	sp<Catalog> spCatalog=m_spOperatorSurfaceI;
	if(!spCatalog.isValid())
	{
		feLog("HoudiniSOP:updateCatalog operator is not a Catalog\n");
		return;
	}

	const int thread=UT_Thread::getMyThreadId();

	PRM_ParmList* pPrmParmList=getParmList();	// as OP_Parameters
	U32 parms=pPrmParmList->getEntries();
	for(U32 parm=0;parm<parms;parm++)
	{
		PRM_Parm* pPrmParm=pPrmParmList->getParmPtr(parm);
		const PRM_Type& rPrmType=pPrmParm->getType();

		const char* token=pPrmParm->getToken();
		const char* label=pPrmParm->getLabel();
		const int size=pPrmParm->getVectorSize();

		//HACK
		String catalogName=String(token).substitute("_"," ");

#if	FE_HSOP_CATALOG_DEBUG
		feLog("HoudiniSOP::updateCatalog \"%s\" \"%s\" %d \"%s\"\n",
				token, label, size, catalogName.c_str());
#endif

		if(rPrmType.isStringType())
		{
			UT_String utString;
			const int index=0;
			evalString(utString,pPrmParm,index,time);

#if	FE_HSOP_CATALOG_DEBUG
			feLog("HoudiniSOP::updateCatalog string \"%s\"\n",
					(char*)utString);

			UT_String utExpression;
			pPrmParm->getExpressionOnly(time,utExpression,0,thread);
			String expression((char*)utExpression);
			feLog("as expression \"%s\"\n",expression.c_str());
#endif

			spCatalog->catalog<String>(catalogName)=utString;
		}
		else if(rPrmType.isFloatType())
		{
			const PRM_Type::PRM_FloatType& prmFloatType=rPrmType.getFloatType();
			if(prmFloatType==PRM_Type::PRM_FLOAT_INTEGER)
			{
				const int index=0;
				const int value=evalInt(pPrmParm,index,time);

#if	FE_HSOP_CATALOG_DEBUG
				feLog("HoudiniSOP::updateCatalog int %d\n",value);
#endif

				spCatalog->catalog<I32>(catalogName)=value;
			}
			else
			{

				float value[size];
				evalFloats(pPrmParm,value,time);

				switch(size)
				{
					case 1:
#if	FE_HSOP_CATALOG_DEBUG
						feLog("HoudiniSOP::updateCatalog float %.6G\n",
								value[0]);
#endif

						spCatalog->catalog<Real>(catalogName)=value[0];
						break;
					case 3:
#if	FE_HSOP_CATALOG_DEBUG
						feLog("HoudiniSOP::updateCatalog float[3]"
								" %.6G %.6G %.6G\n",
								value[0],value[1],value[2]);
#endif

						spCatalog->catalog<SpatialVector>(catalogName)=
								SpatialVector(value);
						break;
					case 4:
#if	FE_HSOP_CATALOG_DEBUG
						feLog("HoudiniSOP::updateCatalog float[4]"
								" %.6G %.6G %.6G %.6G\n",
								value[0],value[1],value[2],value[3]);
#endif

						spCatalog->catalog<Vector4>(catalogName)=Vector4(value);
						break;
					default:
						break;
				}
			}
		}
		else if(rPrmType.isOrdinalType())
		{
			const int index=0;
			const int value=evalInt(pPrmParm,index,time);

#if	FE_HSOP_CATALOG_DEBUG
			feLog("HoudiniSOP::updateCatalog ordinal %d\n",value);
#endif

			spCatalog->catalog<bool>(catalogName)=value;
		}
		else
		{
#if	FE_HSOP_CATALOG_DEBUG
			feLog("HoudiniSOP::updateCatalog unknown type\n");
#endif
		}
	}

	sp<Registry> spRegistry=m_houdiniContext.master()->registry();

	std::vector<String> keys;
	spCatalog->catalogKeys(keys);

	const U32 number=keys.size();
	for(U32 m=0;m<number;m++)
	{
		Instance instance;
		bool success=spCatalog->catalogLookup(keys[m],instance);
		if(!success)
		{
			feLog("HoudiniSOP::updateCatalog invalid Instance \"%s\"\n",
					keys[m].c_str());
			continue;
		}

		char* cleanname=strdup(keys[m].substitute(" ","_").c_str());
		char* keyname=strdup(keys[m].c_str());

		if(instance.is< sp<Component> >())
		{
			sp<Component>& rspComponent=instance.cast< sp<Component> >();

			if(!rspComponent.isValid())
			{
				String implementation=spCatalog->catalogOrDefault<String>(
						keyname,"implementation",
						"SurfaceAccessibleI.SurfaceAccessibleHoudini");
				if(!implementation.empty())
				{
					if(implementation=="SurfaceAccessibleI")
					{
						implementation=
								"SurfaceAccessibleI.SurfaceAccessibleHoudini";
					}
					rspComponent=spRegistry->create(implementation);
				}
				sp<SurfaceAccessibleI> spSurfaceAccessibleI=rspComponent;
				if(spSurfaceAccessibleI.isValid())
				{
					String io=spCatalog->catalogOrDefault<String>(
							keyname,"IO","input");
					if(io=="input")
					{
						I32 channel=m_surfaceInputArray.size();
#if	FE_HSOP_CATALOG_DEBUG
						feLog("HoudiniSOP::updateCatalog"
								" adding surface input %d \"%s\" \"%s\"\n",
								channel,cleanname,keyname);
#endif
						spCatalog->catalog<I32>(keyname,"channel")=channel;
						m_surfaceInputArray.push_back(spSurfaceAccessibleI);
					}
					else
					{
#if	FE_HSOP_CATALOG_DEBUG
						feLog("HoudiniSOP::updateCatalog"
								" adding surface output \"%s\" \"%s\"\n",
								cleanname,keyname);
#endif
						m_nameOutput=keyname;
						m_spSurfaceOutput=spSurfaceAccessibleI;
					}
				}
			}
		}
	}
}

OP_ERROR HoudiniSOP::cookMySop(OP_Context &context)
{
	double time=context.getTime();

	// Before we do anything, we must lock our inputs.  Before returning,
	//	we have to make sure that the inputs get unlocked.
	if (lockInputs(context) >= UT_ERROR_ABORT)
	{
		return error();
	}

	UT_String operatorParam;
	evalString(operatorParam, 0, 0, time);
	String operatorName((char*)operatorParam);
#if	FE_HSOP_COOK_DEBUG
	feLog("--------------------------------\n");
	feLog("HoudiniSOP::cookMySop %f \"%s\"\n",time,operatorName.c_str());
#endif

	const BWORD resetState=(time==0.0f && m_lastTime>0.0);

	if(m_spOperatorSurfaceI.isValid())
	{
//		feLog("HoudiniSOP::cookMySop component name \"%s\" as \"%s\"\n",
//				m_spOperatorSurfaceI->name().c_str(),m_lastName.c_str());
		if(resetState || !m_lastName.dotMatch(operatorName))
		{
			if(m_spSignalerI.isValid())
			{
				m_spSignalerI->remove(m_spOperatorSurfaceI, m_spLayout);
			}
			m_spOperatorSurfaceI=NULL;
			m_lastName="";
			m_surfaceInputArray.clear();
		}
/*
		else if(time<m_lastTime)
		{
			feLog("HoudiniSOP::cookMySop time %f before last %f\n",
					time,m_lastTime);

			unlockInputs();
			return error();
		}
*/
	}
	if(!m_spOperatorSurfaceI.isValid())
	{
		sp<Registry> spRegistry=m_houdiniContext.master()->registry();
		m_spOperatorSurfaceI=spRegistry->create(operatorName);
//		feLog("HoudiniSOP::cookMySop m_spOperatorSurfaceI \"%s\" valid=%d\n",
//				operatorName.c_str(),m_spOperatorSurfaceI.isValid());

		if(m_spOperatorSurfaceI.isValid())
		{
			updateCatalog(time);

			m_lastName=operatorName;

			if(m_spSignalerI.isValid())
			{
				m_spSignalerI->insert(m_spOperatorSurfaceI, m_spLayout);
			}
		}
	}
	else
	{
		updateCatalog(time);
	}

	if(m_spOperatorSurfaceI.isValid())
	{
		// Flag the SOP as being time dependent (i.e. cook on time changes)
		flags().timeDep = 1;

		double frame=OPgetDirector()->getChannelManager()->getSample(time);

		sp<HoudiniDraw> spHoudiniDraw=m_spDrawI;
		if(spHoudiniDraw.isValid())
		{
//			feLog("HoudiniSOP::cookMySop set draw gdp=%p\n",gdp);
			spHoudiniDraw->setGdp(gdp);
		}

		U32 surfaces=m_surfaceInputArray.size();
#if	FE_HSOP_COOK_DEBUG
		feLog("surfaces %d\n",surfaces);
#endif
		for(int surface=0;surface<surfaces;surface++)
		{
			const GU_Detail* pGdp=inputGeo(surface);
			sp<SurfaceAccessibleHoudini> spSurfaceInput=
					m_surfaceInputArray[surface];
			spSurfaceInput->setGdp(pGdp);
		}

		sp<SurfaceAccessibleHoudini> spSurfaceOutput=m_spSurfaceOutput;
		if(!spSurfaceOutput.isValid())
		{
			sp<Registry> spRegistry=m_houdiniContext.master()->registry();
			spSurfaceOutput=spRegistry->create(
					"SurfaceAccessibleI.SurfaceAccessibleHoudini");
		}
//		feLog("HoudiniSOP::cookMySop spSurfaceOutput valid=%d\n",
//				spSurfaceOutput.isValid());

		sp<Catalog> spCatalog=m_spOperatorSurfaceI;

		bool recycle=false;
		I32 channel=0;
		if(!m_nameOutput.empty())
		{
			channel= -1;

			String copy=spCatalog->catalogOrDefault<String>(m_nameOutput,
					"copy","");
			if(!copy.empty())
			{
				channel=spCatalog->catalogOrDefault<I32>(copy,"channel",-1);

				recycle=spCatalog->catalogOrDefault<bool>(m_nameOutput,
						"recycle",false);
			}
		}

		int changed=1;
		if(channel>=0 && channel<surfaces)
		{
			if(recycle && !resetState)
			{
				// reuse copy of input geometry as output
				duplicateChangedSource(channel,context,&changed);
			}
			else
			{
				// repeatedly copy input geometry to output
				duplicateSource(channel,context);
			}
		}
		else
		{
			// erase all prior geometry
			gdp->clearAndDestroy();
		}

#if	FE_HSOP_COOK_DEBUG
		feLog("set gdp %p\n",gdp);
#endif
		spSurfaceOutput->setGdpChangeable(gdp);

		m_aSurfaceOutput(m_signal)=spSurfaceOutput;
		m_spSignalerI->signal(m_signal);

		String errorString=spCatalog->catalogOrDefault<String>("error","");
		if(!errorString.empty())
		{
			addError(SOP_MESSAGE, errorString.c_str());
			spCatalog->catalog<String>("error")="";
		}
#if	FE_HSOP_COOK_DEBUG
		feLog("HoudiniSOP::cookMySop error string \"%s\"\n",
				errorString.c_str());
#endif
	}

	m_lastTime=time;

	unlockInputs();

#if	FE_HSOP_COOK_DEBUG
	feLog("HoudiniSOP::cookMySop done\n");
#endif

	return error();
}

} // namespace
