/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "operator.pmh"

#define	FE_OPC_DEBUG	FALSE

namespace fe
{

FE_DL_PUBLIC sp<Master> OperatorContext::ms_spMaster;

OperatorContext::OperatorContext(void)
{
#if FE_OPC_DEBUG
	feLog("OperatorContext\n");
#endif

	m_spMasterHold=ms_spMaster;

#if FE_OPC_DEBUG
	feLog("OperatorContext done\n");
#endif
}

OperatorContext::~OperatorContext(void)
{
	m_spMasterHold=NULL;
	m_spScope=NULL;

#if FE_OPC_DEBUG
	if(ms_spMaster.isValid())
	{
		feLog("~OperatorContext Master count is %d\n",ms_spMaster->count());
	}
#endif

	if(ms_spMaster.isValid() && ms_spMaster->count()==1)
	{
#if FE_OPC_DEBUG
		feLog("~OperatorContext releasing Master\n");
#endif
		ms_spMaster=NULL;
	}
}

sp<Master> OperatorContext::master(void)
{
#if FE_OPC_DEBUG
	feLog("OperatorContext::master\n");
#endif

	if(!ms_spMaster.isValid())
	{
		ms_spMaster=new Master();
	}

#if FE_OPC_DEBUG
	feLog("OperatorContext::master %p\n",ms_spMaster.raw())
#endif

	m_spMasterHold=ms_spMaster;
	return ms_spMaster;
}

sp<Scope> OperatorContext::scope(void)
{
	if(!m_spScope.isValid() && master().isValid())
	{
		m_spScope=master()->catalog()->catalogComponent(
				"Scope","OperatorScope");

#if FE_OPC_DEBUG
	feLog("OperatorContext::scope 0x%p\n",m_spScope.raw())
#endif
	}
	return m_spScope;
}

void OperatorContext::loadLibraries(void)
{
#if FE_OPC_DEBUG
	feLog("OperatorContext:loadLibraries\n");
#endif

	sp<Registry> spRegistry=master()->registry();

	String libs;
	System::getEnvironmentVariable(m_libEnvVar, libs);
	if(libs.empty())
	{
		System::getEnvironmentVariable("FE_OPERATOR_LIBS", libs);
	}
	if(libs.empty())
	{
		libs="fexOperatorDL";
	}

#if FE_OPC_DEBUG
	feLog("OperatorContext:loadLibraries managing \"%s\"\n",libs.c_str());
#endif

	while(!libs.empty())
	{
		String lib=libs.parse("'",":");
		if(lib.empty())
		{
			break;
		}
		Result result=spRegistry->manage(lib);
#if FE_OPC_DEBUG
		feLog("OperatorContext:loadLibraries manage \"%s\" result=%p\n",
				lib.c_str(),result);
#endif
	}
}

} // namespace
