/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "fe/core.h"
#include "core/Variant.h"

class MyClass
{
	public:
		int x;
};

int main()
{
	UNIT_START();

	F32 fValue=1.23f;
	U32 uValue=4;
	I32 iValue=5;

	fe::Variant fVariant=&fValue;
	fe::Variant uVariant=&uValue;
	fe::Variant iVariant=&iValue;
	fe::Variant vVariant=(void*)&iValue;

	fe::Variant nVariant=0;	// null
	fe::Variant eVariant;	// empty
	fe::Variant cVariant;	// copy

	fVariant=&fValue;
	nVariant=0;
	cVariant=fVariant;

	UNIT_TEST(FE_ISTYPE(F32*,fVariant));
	UNIT_TEST(!FE_ISTYPE(U32*,fVariant));

	UNIT_TEST(fVariant.typeID()==typeid(F32*));

	UNIT_TEST(!fVariant.isEmpty());
	UNIT_TEST(eVariant.isEmpty());

	UNIT_TEST(!fVariant.isNull());
	UNIT_TEST(!uVariant.isNull());
	UNIT_TEST(!iVariant.isNull());
	UNIT_TEST(!vVariant.isNull());
	UNIT_TEST(nVariant.isNull());
	UNIT_TEST(!eVariant.isNull());

	F32 *fPointer=fVariant.castTo<F32*>();
	U32 *uPointer=uVariant.castTo<U32*>();
	I32 *iPointer=iVariant.castTo<I32*>();
	void *vPointer=vVariant.castTo<void*>();

	UNIT_TEST(fPointer);
	UNIT_TEST(uPointer);
	UNIT_TEST(iPointer);
	UNIT_TEST(vPointer);

	UNIT_TEST(fPointer && *fPointer==1.23f);
	UNIT_TEST(uPointer && *uPointer==4);
	UNIT_TEST(iPointer && *iPointer==5);
	UNIT_TEST(vPointer && *(I32*)vPointer==5);

	fPointer=cVariant.castTo<F32*>();
	uPointer=cVariant.castTo<U32*>();

	UNIT_TEST(fPointer);
	UNIT_TEST(fPointer && *fPointer==1.23f);
	UNIT_TEST(!uPointer);

	fPointer=vVariant.castTo<F32*>();
	uPointer=fVariant.castTo<U32*>();
	iPointer=uVariant.castTo<I32*>();
	vPointer=iVariant.castTo<void*>();

	UNIT_TEST(!fPointer);
	UNIT_TEST(!uPointer);
	UNIT_TEST(!iPointer);
	UNIT_TEST(!vPointer);

	UNIT_TRACK(26);
	UNIT_RETURN();
}
